# validated: 2019-01-02 DV ecfe95383cdf edu/wpi/first/wpilibj/SensorUtil.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2018. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

import hal

__all__ = ["SensorUtil"]


class SensorUtil:
    """
    Stores most recent status information as well as containing utility
    functions for checking channels and error processing.
    """

    #: Ticks per microsecond
    kSystemClockTicksPerMicrosecond = hal.getSystemClockTicksPerMicrosecond()

    #: Number of digital channels per roboRIO
    kDigitalChannels = hal.getNumDigitalChannels()

    #: Number of analog input channels per roboRIO
    kAnalogInputChannels = hal.getNumAnalogInputs()

    #: Number of analog output channels per roboRIO
    kAnalogOutputChannels = hal.getNumAnalogOutputs()

    #: Number of solenoid channels per module
    kSolenoidChannels = hal.getNumSolenoidChannels()

    #: Number of PWM channels per roboRIO
    kPwmChannels = hal.getNumPWMChannels()

    #: Number of relay channels per roboRIO
    kRelayChannels = hal.getNumRelayHeaders()

    #: Number of power distribution channels per PDP
    kPDPChannels = hal.getNumPDPChannels()

    #: Number of power distribution modules per PDP
    kPDPModules = hal.getNumPDPModules()

    #: Number of PCM modules
    kPCMModules = hal.getNumPCMModules()

    @staticmethod
    def checkSolenoidModule(moduleNumber: int) -> None:
        """Verify that the solenoid module is correct.

        :param moduleNumber: The solenoid module module number to check.
        """
        if not hal.checkSolenoidModule(moduleNumber):
            raise IndexError(
                "Requested solenoid module number %d is out of range [0, %d)."
                % (moduleNumber, SensorUtil.kPCMModules)
            )

    @staticmethod
    def checkDigitalChannel(channel: int) -> None:
        """Check that the digital channel number is valid.
        Verify that the channel number is one of the legal channel numbers.
        Channel numbers are 0-based.

        :param channel: The channel number to check.
        """
        if not hal.checkDIOChannel(channel):
            raise IndexError(
                "Requested digital channel number %d is out of range [0, %d)."
                % (channel, SensorUtil.kDigitalChannels)
            )

    @staticmethod
    def checkRelayChannel(channel: int) -> None:
        """Check that the digital channel number is valid.
        Verify that the channel number is one of the legal channel numbers.
        Channel numbers are 0-based.

        :param channel: The channel number to check.
        """
        if not hal.checkRelayChannel(channel):
            raise IndexError(
                "Requested relay channel number %d is out of range [0, %d)."
                % (channel, SensorUtil.kRelayChannels)
            )

    @staticmethod
    def checkPWMChannel(channel: int) -> None:
        """Check that the digital channel number is valid.
        Verify that the channel number is one of the legal channel numbers.
        Channel numbers are 0-based.

        :param channel: The channel number to check.
        """
        if not hal.checkPWMChannel(channel):
            raise IndexError(
                "Requested PWM channel number %d is out of range [0, %d)."
                % (channel, SensorUtil.kPwmChannels)
            )

    @staticmethod
    def checkAnalogInputChannel(channel: int) -> None:
        """Check that the analog input number is value.
        Verify that the analog input number is one of the legal channel
        numbers.  Channel numbers are 0-based.

        :param channel: The channel number to check.
        """
        if not hal.checkAnalogInputChannel(channel):
            raise IndexError(
                "Requested analog input channel number %d is out of range [0, %d)."
                % (channel, SensorUtil.kAnalogInputChannels)
            )

    @staticmethod
    def checkAnalogOutputChannel(channel: int) -> None:
        """Check that the analog input number is value.
        Verify that the analog input number is one of the legal channel
        numbers.  Channel numbers are 0-based.

        :param channel: The channel number to check.
        """
        if not hal.checkAnalogOutputChannel(channel):
            raise IndexError(
                "Requested analog output channel number %d is out of range [0, %d)."
                % (channel, SensorUtil.kAnalogOutputChannels)
            )

    @staticmethod
    def checkSolenoidChannel(channel: int) -> None:
        """Verify that the solenoid channel number is within limits.  Channel
        numbers are 0-based.

        :param channel: The channel number to check.
        """
        if not hal.checkSolenoidChannel(channel):
            raise IndexError(
                "Requested solenoid channel number %d is out of range [0, %d)."
                % (channel, SensorUtil.kSolenoidChannels)
            )

    @staticmethod
    def checkPDPChannel(channel: int) -> None:
        """Verify that the power distribution channel number is within limits.
        Channel numbers are 0-based.

        :param channel: The channel number to check.
        """
        if not hal.checkPDPChannel(channel):
            raise IndexError(
                "Requested PDP channel number %d is out of range [0, %d)."
                % (channel, SensorUtil.kPDPChannels)
            )

    @staticmethod
    def checkPDPModule(module: int) -> None:
        """Verify that the power distribution module number is within limits.
        Module numbers are 0-based.

        :param module: The module number to check.
        """
        if not hal.checkPDPModule(module):
            raise IndexError(
                "Requested PDP module number %d is out of range [0, %d)."
                % (module, SensorUtil.kPDPModules)
            )

    @staticmethod
    def getDefaultSolenoidModule() -> int:
        """Get the number of the default solenoid module.

        :returns: The number of the default solenoid module.
        """
        return 0



# validated: 2019-01-02 EN 59700882f1b1 edu/wpi/first/wpilibj/PIDController.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2016. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

import warnings

from .sendablebuilder import SendableBuilder

from .interfaces.pidsource import PIDSource
from .lineardigitalfilter import LinearDigitalFilter
from .pidbase import PIDBase
from .notifier import Notifier
from ._impl.utils import match_arglist, HasAttribute

__all__ = ["PIDController"]


class PIDController(PIDBase):
    """Can be used to control devices via a PID Control Loop.

    Creates a separate thread which reads the given :class:`.PIDSource` and takes
    care of the integral calculations, as well as writing the given
    :class:`.PIDOutput`.

    This feedback controller runs in discrete time, so time deltas are not used
    in the integral and derivative calculations. Therefore, the sample rate affects
    the controller's behavior for a given set of PID constants.
    """

    def __init__(self, Kp, Ki, Kd, *args, **kwargs):
        """Allocate a PID object with the given constants for P, I, D, and F

        Arguments can be structured as follows:

        - Kp, Ki, Kd, Kf, source, output, period
        - Kp, Ki, Kd, source, output, period
        - Kp, Ki, Kd, source, output
        - Kp, Ki, Kd, Kf, source, output

        :param Kp: the proportional coefficient
        :param Ki: the integral coefficient
        :param Kd: the derivative coefficient
        :param Kf: the feed forward term
        :param source: Called to get values
        :param output: Receives the output percentage
        :param period: the loop time for doing calculations. This particularly
            effects calculations of the integral and differential terms.
            The default is 0.05 (50ms).
        """
        f_arg = ("Kf", [float, int])
        source_arg = ("source", [HasAttribute("pidGet"), HasAttribute("__call__")])
        output_arg = ("output", [HasAttribute("pidWrite"), HasAttribute("__call__")])
        period_arg = ("period", [float, int])

        templates = [
            [f_arg, source_arg, output_arg, period_arg],
            [source_arg, output_arg, period_arg],
            [source_arg, output_arg],
            [f_arg, source_arg, output_arg],
        ]

        _, results = match_arglist("PIDController.__init__", args, kwargs, templates)

        Kf = results.pop("Kf", 0.0)  # factor for feedforward term
        output = results.pop("output")
        source = results.pop("source")
        super().__init__(Kp, Ki, Kd, Kf, source, output)

        self.period = results.pop("period", self.kDefaultPeriod)

        self.controlLoop = Notifier(self._calculate)
        self.controlLoop.startPeriodic(self.period)

    def close(self) -> None:
        """Free the PID object"""
        super().close()
        # TODO: is this useful in Python?  Should make TableListener weakref.
        if self.controlLoop is not None:
            self.controlLoop.close()
        with self.mutex:
            self.pidInput = None
            self.pidOutput = None
            self.controlLoop = None

    def enable(self) -> None:
        """Begin running the PIDController."""
        with self.mutex:
            self.enabled = True

    def disable(self) -> None:
        """Stop running the PIDController, this sets the output to zero before
        stopping."""
        with self.pidWriteMutex:
            with self.mutex:
                self.enabled = False
            self.pidOutput(0)

    def setEnabled(self, enable: bool) -> None:
        """Set the enabled state of the PIDController."""
        if enable:
            self.enable()
        else:
            self.disable()

    def isEnabled(self) -> bool:
        """Return True if PIDController is enabled."""
        with self.mutex:
            return self.enabled

    def reset(self) -> None:
        """Reset the previous error, the integral term, and disable the
        controller."""
        self.disable()
        super().reset()

    def initSendable(self, builder: SendableBuilder) -> None:
        super().initSendable(builder)
        builder.addBooleanProperty("enabled", self.isEnabled, self.setEnabled)



# validated: 2019-01-02 DV f0f196e5b389 edu/wpi/first/wpilibj/MotorSafety.java
# ----------------------------------------------------------------------------
# Copyright (c) 2008-2018 FIRST. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

import logging
import threading
import weakref

from .timer import Timer

logger = logging.getLogger(__name__)

__all__ = ["MotorSafety"]


class MotorSafety:
    """Provides mechanisms to safely shutdown motors if they aren't updated
    often enough.

    This base class runs a watchdog timer and calls the subclass's stopMotor()
    function if the timeout expires.

    The subclass should call feed() whenever the motor value is updated.
    """

    DEFAULT_SAFETY_EXPIRATION = 0.1
    __instanceList = weakref.WeakSet()
    __listMutex = threading.Lock()

    @classmethod
    def _reset(cls) -> None:
        with cls.__listMutex:
            cls.__instanceList.clear()

    def __init__(self) -> None:
        self.__enabled = False
        self.__expiration = self.DEFAULT_SAFETY_EXPIRATION
        self.__stopTime = Timer.getFPGATimestamp()
        self.__mutex = threading.Lock()
        with self.__listMutex:
            self.__instanceList.add(self)

    def feed(self) -> None:
        """Feed the motor safety object.

        Resets the timer on this object that is used to do the timeouts.
        """
        with self.__mutex:
            self.__stopTime = Timer.getFPGATimestamp() + self.__expiration

    def setExpiration(self, expirationTime: float) -> None:
        """Set the expiration time for the corresponding motor safety object.

        :param expirationTime: The timeout value in seconds.
        """
        with self.__mutex:
            self.__expiration = expirationTime

    def getExpiration(self) -> float:
        """Retrieve the timeout value for the corresponding motor safety object.

        :returns: the timeout value in seconds.
        """
        with self.__mutex:
            return self.__expiration

    def isAlive(self) -> bool:
        """Determine of the motor is still operating or has timed out.

        :returns: True if the motor is still operating normally and hasn't
            timed out.
        """
        with self.__mutex:
            return not self.__enabled or self.__stopTime > Timer.getFPGATimestamp()

    def check(self) -> None:
        """Check if this motor has exceeded its timeout.
        This method is called periodically to determine if this motor has
        exceeded its timeout value. If it has, the stop method is called,
        and the motor is shut down until its value is updated again.
        """
        from .robotstate import RobotState

        with self.__mutex:
            enabled = self.__enabled
            stopTime = self.__stopTime

        if not enabled or RobotState.isDisabled() or RobotState.isTest():
            return

        if stopTime < Timer.getFPGATimestamp():
            # TODO: fix this, causes recursion error
            logger.warning(
                "%s... Output not updated often enough." % self.getDescription()
            )

            self.stopMotor()

    def setSafetyEnabled(self, enabled: bool) -> None:
        """Enable/disable motor safety for this device.

        Turn on and off the motor safety option for this PWM object.

        :param enabled: True if motor safety is enforced for this object
        """
        with self.__mutex:
            self.__enabled = bool(enabled)

    def isSafetyEnabled(self) -> bool:
        """Return the state of the motor safety enabled flag.

        Return if the motor safety is currently enabled for this device.

        :returns: True if motor safety is enforced for this device
        """
        with self.__mutex:
            return self.__enabled

    @classmethod
    def checkMotors(cls) -> None:
        """Check the motors to see if any have timed out.
        This static method is called periodically to poll all the motors and
        stop any that have timed out.
        """
        with cls.__listMutex:
            for elem in cls.__instanceList:
                elem.check()

    # abstract methods

    def stopMotor(self) -> None:
        raise NotImplementedError

    def getDescription(self) -> str:
        raise NotImplementedError



from .shuffleboard import Shuffleboard
from .complexwidget import ComplexWidget
from .component import ShuffleboardComponent
from .container import ShuffleboardContainer
from .eventimportance import EventImportance
from .layout import ShuffleboardLayout
from .simplewidget import SimpleWidget
from .tab import ShuffleboardTab



# validated: 2018-09-09 EN ecfe95383cdf edu/wpi/first/wpilibj/AnalogPotentiometer.java
# ----------------------------------------------------------------------------*/
# Copyright (c) FIRST 2008-2017. All Rights Reserved.                        */
# Open Source Software - may be modified and shared by FRC teams. The code   */
# must be accompanied by the FIRST BSD license file in the root directory of */
# the project.                                                               */
# ----------------------------------------------------------------------------*/
from typing import Union

import hal

from .analoginput import AnalogInput
from .interfaces import PIDSource
from .sendablebase import SendableBase
from .sendablebuilder import SendableBuilder

__all__ = ["AnalogPotentiometer"]


class AnalogPotentiometer(SendableBase):
    """Reads a potentiometer via an :class:`.AnalogInput`
    
    Analog potentiometers read
    in an analog voltage that corresponds to a position. The position is in
    whichever units you choose, by way of the scaling and offset constants
    passed to the constructor.

    .. not_implemented: initPot
    """

    PIDSourceType = PIDSource.PIDSourceType

    def __init__(
        self,
        channel: Union[AnalogInput, int],
        fullRange: float = 1.0,
        offset: float = 0.0,
    ) -> None:
        """AnalogPotentiometer constructor.

        Use the fullRange and offset values so that the output produces
        meaningful values. I.E: you have a 270 degree potentiometer and
        you want the output to be degrees with the halfway point as 0
        degrees. The fullRange value is 270.0(degrees) and the offset is
        -135.0 since the halfway point after scaling is 135 degrees.

        :param channel: The analog channel this potentiometer is plugged into.
        :param fullRange: The scaling to multiply the fraction by to get a
            meaningful unit.  Defaults to 1.0 if unspecified.
        :param offset: The offset to add to the scaled value for controlling
            the zero value.  Defaults to 0.0 if unspecified.
        """

        super().__init__()
        if not hasattr(channel, "getVoltage"):
            channel = AnalogInput(channel)
            self.addChild(channel)
        self.analog_input = channel
        self.fullRange = fullRange
        self.offset = offset
        self.init_analog_input = True
        self.pidSource = self.PIDSourceType.kDisplacement
        self.setName("AnalogPotentiometer", self.analog_input.getChannel())

    def get(self) -> float:
        """Get the current reading of the potentiometer.

        :returns: The current position of the potentiometer.
        """
        if self.analog_input is None:
            return self.offset
        return (
            self.analog_input.getVoltage() / hal.getUserVoltage5V()
        ) * self.fullRange + self.offset

    def setPIDSourceType(self, pidSource: PIDSourceType) -> None:
        """Set which parameter you are using as a process
        control variable. 

        :param pidSource: An enum to select the parameter.
        """
        if pidSource != self.PIDSourceType.kDisplacement:
            raise ValueError("Only displacement PID is allowed for potentiometers.")
        self.pidSource = pidSource

    def getPIDSourceType(self) -> PIDSourceType:
        return self.pidSource

    def pidGet(self) -> float:
        """Implement the PIDSource interface.

        :returns: The current reading.
        """
        return self.get()

    def initSendable(self, builder: SendableBuilder) -> None:
        if self.analog_input is not None:
            self.analog_input.initSendable(builder)

    def close(self) -> None:
        super().close()

        if self.init_analog_input:
            self.analog_input.close()
            del self.analog_input
            self.init_analog_input = False



# validated: 2018-09-09 EN ecfe95383cdf edu/wpi/first/wpilibj/GyroBase.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

from .interfaces import PIDSource
from .sendablebase import SendableBase
from .sendablebuilder import SendableBuilder

__all__ = ["GyroBase"]


class GyroBase(SendableBase):
    """
        GyroBase is the common base class for Gyro implementations such as
        :class:`.AnalogGyro`.
    """

    PIDSourceType = PIDSource.PIDSourceType

    def __init__(self) -> None:
        super().__init__()
        self.pidSource = self.PIDSourceType.kDisplacement
        self.valueEntry = None

    def calibrate(self) -> None:
        raise NotImplementedError()

    def reset(self) -> None:
        raise NotImplementedError()

    def getAngle(self) -> float:
        raise NotImplementedError()

    def getRate(self) -> float:
        raise NotImplementedError()

    def setPIDSourceType(self, pidSource: PIDSourceType) -> None:
        """Set which parameter of the gyro you are using as a process
        control variable. The Gyro class supports the rate and angle
        parameters.

        :param pidSource: An enum to select the parameter.
        """
        if pidSource not in (
            self.PIDSourceType.kDisplacement,
            self.PIDSourceType.kRate,
        ):
            raise ValueError("Must be kRate or kDisplacement")
        self.pidSource = pidSource

    def getPIDSourceType(self) -> PIDSourceType:
        return self.pidSource

    def pidGet(self) -> float:
        """Get the output of the gyro for use with PIDControllers. May be
        the angle or rate depending on the set :class:`.PIDSourceType`

        :returns: the current angle according to the gyro
        """
        if self.pidSource == self.PIDSourceType.kRate:
            return self.getRate()
        elif self.pidSource == self.PIDSourceType.kDisplacement:
            return self.getAngle()
        else:
            return 0.0

    def initSendable(self, builder: SendableBuilder) -> None:
        builder.setSmartDashboardType("Gyro")
        builder.addDoubleProperty("Value", self.getAngle, None)



# validated: 2018-09-09 EN ecfe95383cdf edu/wpi/first/wpilibj/DigitalOutput.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

import warnings
import weakref
from typing import Optional

import hal

from .digitalsource import DigitalSource
from .sensorutil import SensorUtil
from .sendablebase import SendableBase
from .sendablebuilder import SendableBuilder

__all__ = ["DigitalOutput"]


def _freePWMGenerator(pwmGenerator: int) -> None:
    # Disable the output by routing to a dead bit.
    hal.setDigitalPWMOutputChannel(pwmGenerator, SensorUtil.kDigitalChannels)
    hal.freeDigitalPWM(pwmGenerator)


class DigitalOutput(SendableBase):
    """Writes to a digital output
    
    Other devices that are implemented elsewhere will automatically allocate
    digital inputs and outputs as required.
    """

    invalidPwmGenerator = None

    def __init__(self, channel: int) -> None:
        """Create an instance of a digital output.

        :param channel: the DIO channel for the digital output. 0-9 are on-board, 10-25 are on the MXP
        """

        super().__init__()
        self._pwmGenerator = None
        self._pwmGenerator_finalizer = None

        SensorUtil.checkDigitalChannel(channel)
        self.channel = channel

        self.handle = hal.initializeDIOPort(hal.getPort(channel), False)

        hal.report(hal.UsageReporting.kResourceType_DigitalOutput, channel)
        self.setName("DigitalOutput", channel)

    @property
    def pwmGenerator(self) -> Optional[int]:
        if self._pwmGenerator_finalizer is None:
            return None
        if not self._pwmGenerator_finalizer.alive:
            return None
        return self._pwmGenerator

    def close(self) -> None:
        """Free the resources associated with a digital output."""
        super().close()
        # finalize the pwm only if we have allocated it
        if self.pwmGenerator is not None:
            self._pwmGenerator_finalizer()
        self._pwmGenerator = None
        if self.pwmGenerator is not self.invalidPwmGenerator:
            self.disablePWM()
        hal.freeDIOPort(self.handle)
        self.handle = 0

    def set(self, value: bool) -> None:
        """Set the value of a digital output.

        :param value: True is on, off is False
        """
        hal.setDIO(self.handle, bool(value))

    def get(self) -> bool:
        """Gets the value being output from the Digital Output.

        :returns: the state of the digital output
        """
        return hal.getDIO(self.handle)

    def getChannel(self) -> int:
        """:returns: The GPIO channel number that this object represents.
        """
        return self.channel

    def pulse(self, pulseLength: float) -> None:
        """Generate a single pulse. There can only be a single pulse going at any time.

        :param pulseLength: The length of the pulse.
        """
        hal.pulse(self.handle, pulseLength)

    def isPulsing(self) -> bool:
        """Determine if the pulse is still going. Determine if a previously
        started pulse is still going.

        :returns: True if pulsing
        """
        return hal.isPulsing(self.handle)

    def setPWMRate(self, rate: float) -> None:
        """Change the PWM frequency of the PWM output on a Digital Output line.

        The valid range is from 0.6 Hz to 19 kHz. The frequency resolution is
        logarithmic.

        There is only one PWM frequency for all channels.

        :param rate: The frequency to output all digital output PWM signals.
        """
        hal.setDigitalPWMRate(rate)

    def enablePWM(self, initialDutyCycle: float) -> None:
        """Enable a PWM Output on this line.

        Allocate one of the 6 DO PWM generator resources.

        Supply the initial duty-cycle to output so as to avoid a glitch when
        first starting.

        The resolution of the duty cycle is 8-bit for low frequencies (1kHz or
        less) but is reduced the higher the frequency of the PWM signal is.

        :param initialDutyCycle: The duty-cycle to start generating. [0..1]
        """
        if self.pwmGenerator is not self.invalidPwmGenerator:
            return
        self._pwmGenerator = hal.allocateDigitalPWM()
        hal.setDigitalPWMDutyCycle(self._pwmGenerator, initialDutyCycle)
        hal.setDigitalPWMOutputChannel(self._pwmGenerator, self.channel)
        self._pwmGenerator_finalizer = weakref.finalize(
            self, _freePWMGenerator, self._pwmGenerator
        )

    def disablePWM(self) -> None:
        """Change this line from a PWM output back to a static Digital Output
        line.

        Free up one of the 6 DO PWM generator resources that were in use.
        """
        if self.pwmGenerator is not self.invalidPwmGenerator:
            return
        hal.setDigitalPWMOutputChannel(self._pwmGenerator, SensorUtil.kDigitalChannels)
        hal.freeDigitalPWM(self._pwmGenerator)
        self._pwmGenerator_finalizer()

    def updateDutyCycle(self, dutyCycle: float) -> None:
        """Change the duty-cycle that is being generated on the line.

        The resolution of the duty cycle is 8-bit for low frequencies (1kHz or
        less) but is reduced the higher the frequency of the PWM signal is.

        :param dutyCycle: The duty-cycle to change to. [0..1]
        """
        if self.pwmGenerator is self.invalidPwmGenerator:
            return
        hal.setDigitalPWMDutyCycle(self._pwmGenerator, dutyCycle)

    def initSendable(self, builder: SendableBuilder) -> None:
        builder.setSmartDashboardType("Digital Output")
        builder.addBooleanProperty("Value", self.get, self.set)



# validated: 2017-10-03 EN e1195e8b9dab edu/wpi/first/wpilibj/command/WaitCommand.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2016. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------
from typing import Optional

from .timedcommand import TimedCommand

__all__ = ["WaitCommand"]


class WaitCommand(TimedCommand):
    """A WaitCommand will wait for a certain amount of time before finishing.
    It is useful if you want a :class:`.CommandGroup` to pause for a moment.
    
    .. seealso:: :class:`.CommandGroup`
    """

    def __init__(self, timeout: float, name: Optional[str] = None) -> None:
        """Instantiates a WaitCommand with the given timeout.
        
        :param timeout: the time the command takes to run (seconds)
        :param name: the name of the command (optional)
        """
        if name is None:
            super().__init__("Wait(%s)" % timeout, timeout)
        else:
            super().__init__(name, timeout)



from .differentialdrive import DifferentialDrive
from .killoughdrive import KilloughDrive
from .mecanumdrive import MecanumDrive
from .robotdrivebase import RobotDriveBase
from .vector2d import Vector2d



# validated: 2018-12-15 EN 6f0c185a05c9 edu/wpi/first/wpilibj/shuffleboard/Shuffleboard.java
# ----------------------------------------------------------------------------
# Copyright (c) 2018 FIRST. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

from networktables import NetworkTablesInstance
from .tab import ShuffleboardTab
from .instance import ShuffleboardInstance
from .recordingcontroller import RecordingController
from .eventimportance import EventImportance


__all__ = ["Shuffleboard"]


class Shuffleboard:
    """
    The Shuffleboard class provides a mechanism with which data can be added
    and laid out in the Shuffleboard dashboard application from a robot program.
    Tabs and layouts can be specified, as well as choosing which widgets to
    display with and setting properties of these widgets; for example,
    programmers can specify a specific boolean value to be displayed
    with a toggle button instead of the default colored box, or set custom
    colors for that box.

    For example, displaying a boolean entry with a toggle button::

        myBoolean = (Shuffleboard.getTab("Example Tab")
            .add(title="My Boolean", value=False)
            .withWidget("Toggle Button")
            .getEntry())


    Changing the colors of the boolean box::

        myBoolean = (Shuffleboard.getTab("Example Tab")
            .add(title="My Boolean", value=False)
            .withWidget("Boolean Box")
            .withProperties({"colorWhenTrue": "green", "colorWhenFalse": "maroon"})
            .getEntry())

    Specifying a parent layout. Note that the layout type must *always* be
    specified, even if the layout has already been generated by a previously
    defined entry.

    ::

        myBoolean = (Shuffleboard.getTab("Example Tab")
            .getLayout("List", "Example List")
            .add(title="My Boolean", value=False)
            .withWidget("Toggle Button")
            .getEntry())

    Teams are encouraged to set up shuffleboard layouts at the start of the
    robot program.

    """

    #: The name of the base NetworkTable into which all Shuffleboard data will be added.
    kBaseTableName = "/Shuffleboard"

    _root = None
    _recordingController = None

    @classmethod
    def root(cls):
        if cls._root is None:
            cls._root = ShuffleboardInstance(NetworkTablesInstance.getDefault())

        return cls._root

    @classmethod
    def recordingController(cls):
        if cls._recordingController is None:
            cls._recordingController = RecordingController(
                NetworkTablesInstance.getDefault()
            )

        return cls._recordingController

    @classmethod
    def _reset(cls):
        cls._root = None
        cls._recordingController = None

    @classmethod
    def update(cls) -> None:
        """
        Updates all the values in Shuffleboard. Iterative and timed robots are
        pre-configured to call this method in the main robot loop; teams using
        custom robot base classes, or subclass SampleRobot, should make sure
        to call this repeatedly to keep data on the dashboard up to date.
        """
        cls.root().update()

    @classmethod
    def getTab(cls, title: str) -> ShuffleboardTab:
        """
        Gets the Shuffleboard tab with the given title, creating it if it does
        not already exist.

        :param title: the title of the tab

        :returns: the tab with the given title
        """
        return cls.root().getTab(title)

    @classmethod
    def selectTab(cls, index_or_title):
        """
        Selects the tab in the dashboard with the given index in the 
        range [0..n-1], where n is the number of tabs in the dashboard at the 
        time this method is called.
        
        Or

        Selects the tab in the dashboard with the given title.

        :param index_or_title: when an integer, the index of the tab to select. 
                               when a string, the title of the tab to select.
        """
        cls.root().selectTab(index_or_title)

    @classmethod
    def enableActuatorWidgets(cls) -> None:
        """
        Enables user control of widgets containing actuators: speed
        controllers, relays, etc. This should only be used when the robot is
        in test mode. IterativeRobotBase and SampleRobot are both configured to
        call this method when entering test mode; most users should not need to
        use this method directly.
        """
        cls.root().enableActuatorWidgets()

    @classmethod
    def disableActuatorWidgets(cls) -> None:
        """
        Disables user control of widgets containing actuators. For safety
        reasons, actuators should only be controlled while in test mode.
        IterativeRobotBase and SampleRobot are both configured to call this
        method when exiting in test mode; most users should not need to use
        this method directly.
        """
        cls.update()
        cls.root().disableActuatorWidgets()

    @classmethod
    def startRecording(cls):
        """
        Starts data recording on the dashboard. Has no effect if recording 
        is already in progress.

        see :meth:`.stopRecording`
        """
        cls.recordingController().startRecording()

    @classmethod
    def stopRecording(cls):
        """
        Stops data recording on the dashboard. Has no effect if no recording 
        is in progress.

        see :meth:`.startRecording`
        """
        cls.recordingController().stopRecording()

    @classmethod
    def setRecordingFileNameFormat(cls, format: str):
        """
        Sets the file name format for new recording files to use. If recording 
        is in progress when this method is called, it will continue to use the 
        same file. New recordings will use the format.

        To avoid recording files overwriting each other, make sure to use 
        unique recording file names. File name formats accept templates for 
        inserting the date and time when the recording started with the 
        `${date}` and `${time}` templates, respectively. For example,
        the default format is `"recording-${time}"` and recording files 
        created with it will have names like `"recording-2018.01.15.sbr"`. 
        Users are strongly recommended to use the `${time}` template to ensure 
        unique file names.

        :param format: the format for the
                       see :meth:`.clearRecordingFileNameFormat`
        """
        cls.recordingController().setRecordingFileNameFormat(format)

    @classmethod
    def clearRecordingFileNameFormat(cls):
        """
        Clears the custom name format for recording files. 
        New recordings will use the default format.

        see :meth:`.setRecordingFileNameFormat`
        """
        cls.recordingController().clearRecordingFileNameFormat()

    @classmethod
    def addEventMarker(
        cls, name: str, importance: EventImportance, description: str = ""
    ):
        """
        Notifies Shuffleboard of an event. Events can range from as trivial as 
        a change in a command state to as critical as a total power loss or 
        component failure. If Shuffleboard is recording, the event will also 
        be recorded.

        If `name` is `None` or empty, or `importance` is `None`, then
        no event will be sent and an error will be printed to the driver 
        station.

        :param name:        the name of the event
        :param description: a description of the event
        :param importance:  the importance of the event
        """
        cls.recordingController().addEventMarker(name, description, importance)



# validated: 2018-12-15 EN 6f0c185a05c9 edu/wpi/first/wpilibj/shuffleboard/ShuffleboardInstance.java
# ----------------------------------------------------------------------------
# Copyright (c) 2018 FIRST. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------
from .tab import ShuffleboardTab
from .root import ShuffleboardRoot
from .complexwidget import ComplexWidget
from .container import ShuffleboardContainer


class ShuffleboardInstance(ShuffleboardRoot):
    def __init__(self, ntInstance) -> None:
        from .shuffleboard import Shuffleboard

        self.rootTable = ntInstance.getTable(Shuffleboard.kBaseTableName)
        self.rootMetaTable = self.rootTable.getSubTable(".metadata")
        self.selectedTabEntry = self.rootMetaTable.getEntry("Selected")
        self.tabsChanged = False
        self.tabs = {}

    def getTab(self, title: str) -> "ShuffleboardTab":
        if title not in self.tabs:
            self.tabs[title] = ShuffleboardTab(self, title)
            self.tabsChanged = True
        return self.tabs[title]

    def update(self):
        if self.tabsChanged:
            tabTitles = [tab.getTitle() for tab in self.tabs.values()]
            self.rootMetaTable.getEntry("Tabs").forceSetStringArray(tabTitles)
            self.tabsChanged = False

        for tab in self.tabs.values():
            title = tab.getTitle()
            tab.buildInto(self.rootTable, self.rootMetaTable.getSubTable(title))

    def enableActuatorWidgets(self):
        self._applyToAllComplexWidgets(lambda c: c.enableIfActuator())

    def disableActuatorWidgets(self):
        self._applyToAllComplexWidgets(lambda c: c.disableIfActuator())

    def selectTab(self, index_or_title):
        if isinstance(index_or_title, int):
            self.selectedTabEntry.forceSetDouble(index_or_title)
        else:
            self.selectedTabEntry.forceSetString(index_or_title)

    def _applyToAllComplexWidgets(self, func):
        """
        Applies the function `func` to all complex widgets in this root, 
        regardless of how they are nested.

        :param func: the function to apply to all complex widgets
        """
        for tab in self.tabs.values():
            self._apply(tab, func)

    def _apply(self, container, func):
        """
        Applies the function `func` to all complex widgets in `container`. 
        Helper method for :class:`._applyToAllComplexWidgets`.

        """
        for component in container.getComponents():
            if isinstance(component, ComplexWidget):
                func(component)
            if isinstance(component, ShuffleboardContainer):
                self._apply(component, func)



# validated: 2018-11-18 EN 175c6c1f0130 edu/wpi/first/wpilibj/shuffleboard/ComplexWidget.java
# ----------------------------------------------------------------------------
# Copyright (c) 2018 FIRST. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------
from ..sendablebuilder import SendableBuilder
from ..sendable import Sendable
from .widget import ShuffleboardWidget
from .container import ShuffleboardContainer


class ComplexWidget(ShuffleboardWidget):
    def __init__(self, parent: ShuffleboardContainer, title: str, sendable: Sendable):
        super().__init__(parent, title)
        self.sendable = sendable
        self.builder = None

    def buildInto(self, parentTable, metaTable) -> None:
        self.buildMetadata(metaTable)
        if self.builder is None:
            self.builder = SendableBuilder()
            self.builder.setTable(parentTable.getSubTable(self.getTitle()))
            self.sendable.initSendable(self.builder)
            self.builder.startListeners()

        self.builder.updateTable()

    def enableIfActuator(self):
        """
        Enables user control of this widget in the Shuffleboard application. 
        Has no effect if the sendable is not marked as an actuator 
        with :meth:`.SendableBuilder.setActuator`.
        """
        if self.builder.isActuator():
            self.builder.startLiveWindowMode()

    def disableIfActuator(self):
        """
        Disables user control of this widget in the Shuffleboard application. 
        Has no effect if the sendable is not marked as an actuator 
        with :meth:`.SendableBuilder.setActuator`.
        """
        if self.builder.isActuator():
            self.builder.stopLiveWindowMode()



# validated: 2019-01-02 DV a2368a6199b1 edu/wpi/first/wpilibj/Watchdog.java
# ----------------------------------------------------------------------------
# Copyright (c) 2018 FIRST. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

import heapq
import threading
from typing import Callable

import hal

import logging

logger = logging.getLogger("robotpy")


__all__ = ["Watchdog"]


class Watchdog:
    """A class that's a wrapper around a watchdog timer.

    When the timer expires, a message is printed to the console and an optional user-provided
    callback is invoked.

    The watchdog is initialized disabled, so the user needs to call enable() before use.
    """

    # Used for timeout print rate-limiting
    kMinPrintPeriod = 1000000  # us

    _watchdogs = []  # type: List[Watchdog]
    _queueMutex = threading.Lock()
    _schedulerWaiter = threading.Condition(_queueMutex)
    _keepAlive = True
    _thread = None

    @classmethod
    def _reset(cls) -> None:
        with cls._queueMutex:
            thread = cls._thread
        if thread is not None:
            cls._keepAlive = False
            with cls._queueMutex:
                cls._watchdogs.clear()
                cls._schedulerWaiter.notify_all()
                cls._thread = None
            thread.join()
        cls._keepAlive = True

    def __init__(self, timeout: float, callback: Callable[[], None]) -> None:
        """Watchdog constructor.

        :param timeout: The watchdog's timeout in seconds with microsecond resolution.
        :param callback: This function is called when the timeout expires.
        """
        self._startTime = 0  # us
        self._timeout = int(timeout * 1e6)  # us
        self._expirationTime = 0  # us
        self._callback = callback
        self._lastTimeoutPrintTime = 0  # us
        self._lastEpochsPrintTime = 0  # us
        self._epochs = {}  # type: Dict[str, int]
        self._isExpired = False

        #: Enable or disable suppression of the generic timeout message.
        #:
        #: This may be desirable if the user-provided callback already
        #: prints a more specific message.
        self.suppressTimeoutMessage = False

        with self._queueMutex:
            if Watchdog._thread is None:
                Watchdog._thread = threading.Thread(
                    target=Watchdog._schedulerFunc, daemon=True
                )
                Watchdog._thread.start()

    # Elements with sooner expiration times are sorted as lesser.
    # Python's heap queue is a min-heap.

    def __lt__(self, other: "Watchdog") -> bool:
        if self.__class__ is not other.__class__:
            return NotImplemented
        return self._expirationTime < other._expirationTime

    def __eq__(self, other) -> bool:
        return (
            self.__class__ is other.__class__
            and self._expirationTime == other._expirationTime
        )

    def __gt__(self, other: "Watchdog") -> bool:
        if self.__class__ is not other.__class__:
            return NotImplemented
        return self._expirationTime > other._expirationTime

    def getTime(self) -> float:
        """Returns the time in seconds since the watchdog was last fed."""
        return (hal.getFPGATime() - self._startTime) / 1e6

    def setTimeout(self, timeout: float) -> None:
        """Sets the watchdog's timeout.

        :param timeout: The watchdog's timeout in seconds with microsecond
                        resolution.
        """
        self._startTime = hal.getFPGATime()
        self._epochs.clear()
        timeout = int(timeout * 1e6)  # us

        with self._queueMutex:
            self._timeout = timeout
            self._isExpired = False

            watchdogs = self._watchdogs
            try:
                watchdogs.remove(self)
            except ValueError:
                pass
            else:
                heapq.heapify(watchdogs)

            self._expirationTime = self._startTime + timeout
            heapq.heappush(watchdogs, self)
            self._schedulerWaiter.notify_all()

    def getTimeout(self) -> float:
        """Returns the watchdog's timeout in seconds."""
        with self._queueMutex:
            return self._timeout / 1e6

    def isExpired(self) -> bool:
        """Returns true if the watchdog timer has expired."""
        with self._queueMutex:
            return self._isExpired

    def addEpoch(self, epochName: str) -> None:
        """
        Adds time since last epoch to the list printed by printEpochs().

        Epochs are a way to partition the time elapsed so that when
        overruns occur, one can determine which parts of an operation
        consumed the most time.

        :param epochName: The name to associate with the epoch.
        """
        currentTime = hal.getFPGATime()
        self._epochs[epochName] = currentTime - self._startTime
        self._startTime = currentTime

    def printEpochs(self) -> None:
        """Prints list of epochs added so far and their times."""
        now = hal.getFPGATime()
        if now - self._lastEpochsPrintTime > self.kMinPrintPeriod:
            self._lastEpochsPrintTime = now
            for key, value in self._epochs.items():
                logger.info("\t%s: %.6fs", key, value / 1e6)

    def reset(self) -> None:
        """Resets the watchdog timer.

        This also enables the timer if it was previously disabled.
        """
        self.enable()

    def enable(self) -> None:
        """Enables the watchdog timer."""
        self._startTime = hal.getFPGATime()
        self._epochs.clear()

        with self._queueMutex:
            self._isExpired = False

            watchdogs = self._watchdogs
            try:
                watchdogs.remove(self)
            except ValueError:
                pass
            else:
                heapq.heapify(watchdogs)

            self._expirationTime = self._startTime + self._timeout
            heapq.heappush(watchdogs, self)
            self._schedulerWaiter.notify_all()

    def disable(self) -> None:
        """Disables the watchdog timer."""
        with self._queueMutex:
            self._isExpired = False

            watchdogs = self._watchdogs
            try:
                watchdogs.remove(self)
            except ValueError:
                pass
            else:
                heapq.heapify(watchdogs)
            self._schedulerWaiter.notify_all()

    @classmethod
    def _schedulerFunc(cls) -> None:
        # Grab a bunch of things before the loop to avoid some lookups.
        getFPGATime = hal.getFPGATime
        heappop = heapq.heappop
        log = logger.info
        lock = cls._queueMutex
        watchdogs = cls._watchdogs
        cond = cls._schedulerWaiter

        with lock:
            # Python-specific: need a way to terminate the thread.
            while cls._keepAlive:
                if watchdogs:
                    delta = watchdogs[0]._expirationTime - getFPGATime()
                    timedOut = not cond.wait(delta / 1e6)

                    if timedOut:
                        if (
                            not watchdogs
                            or watchdogs[0]._expirationTime > getFPGATime()
                        ):
                            continue

                        # If the condition variable timed out, that means a Watchdog
                        # timeout has occurred, so call its timeout function.
                        watchdog = heappop(watchdogs)

                        now = getFPGATime()
                        if now - watchdog._lastTimeoutPrintTime > cls.kMinPrintPeriod:
                            watchdog._lastTimeoutPrintTime = now
                            if not watchdog.suppressTimeoutMessage:
                                log(
                                    "Watchdog not fed after %.6fs",
                                    watchdog._timeout / 1e6,
                                )
                        lock.release()
                        try:
                            watchdog._callback()
                        except Exception:
                            logger.exception("Uncaught exception in Watchdog callback")
                        lock.acquire()
                        watchdog._isExpired = True

                    # Otherwise, a Watchdog removed itself from the queue (it
                    # notifies the scheduler of this), we are being reset (by
                    # the test harness) or a spurious wakeup occurred, so rewait
                    # with the soonest watchdog timeout.
                else:
                    while cls._keepAlive and not watchdogs:
                        cond.wait()



# validated: 2018-12-16 EN 45f4472d4224 edu/wpi/first/wpilibj/shuffleboard/EventImportance.java
# ----------------------------------------------------------------------------
# Copyright (c) 2018 FIRST. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------
import enum


__all__ = ["EventImportance"]


class EventImportance(enum.Enum):
    """
    The importance of an event marker in Shuffleboard.  The exact meaning 
    of each importance level is up for interpretation on a team-to-team basis, 
    but users should follow the general guidelines of the various importance 
    levels.  The examples given are for reference and may be ignored or 
    considered to be more or less important from team to team.
    """

    #: A trivial event such as a change in command state.
    kTrivial = "TRIVIAL"

    #: A low importance event such as acquisition of a game piece.
    kLow = "LOW"

    #: A "normal" importance event, such as a transition from autonomous mode to teleoperated control.
    kNormal = "NORMAL"

    #: A high-importance event such as scoring a game piece.
    kHigh = "HIGH"

    #: A critically important event such as a brownout, component failure, or software deadlock.
    kCritical = "CRITICAL"



# validated: 2018-09-09 EN d54c2665dc54 edu/wpi/first/wpilibj/DigitalGlitchFilter.java
# ----------------------------------------------------------------------------
# Copyright (c) 2015-2017 FIRST. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in $(WIND_BASE)/WPILib.
# ----------------------------------------------------------------------------
from typing import Union

import hal
import threading

from .digitalsource import DigitalSource
from .encoder import Encoder
from .counter import Counter
from .sendablebuilder import SendableBuilder
from .sensorutil import SensorUtil
from .sendablebase import SendableBase

__all__ = ["DigitalGlitchFilter"]


class DigitalGlitchFilter(SendableBase):
    """
    Class to enable glitch filtering on a set of digital inputs.
    This class will manage adding and removing digital inputs from a FPGA glitch
    filter. The filter lets the user configure the time that an input must remain
    high or low before it is classified as high or low.
    """

    mutex = threading.Lock()
    filterAllocated = [False] * 3

    def __init__(self) -> None:
        super().__init__()
        self.channelIndex = -1
        with self.mutex:
            for i, v in enumerate(self.filterAllocated):
                if not v:
                    self.channelIndex = i
                    self.filterAllocated[i] = True
                    hal.report(
                        hal.UsageReporting.kResourceType_DigitalGlitchFilter,
                        self.channelIndex,
                        0,
                    )
                    self.setName("DigitalGlitchFilter", i)
                    break
            else:
                raise ValueError("No more filters available")

    def close(self) -> None:
        super().close()
        if self.channelIndex >= 0:
            with self.mutex:
                self.filterAllocated[self.channelIndex] = False

            self.channelIndex = -1

    @staticmethod
    def _setFilter(input: DigitalSource, channelIndex: int) -> None:
        if input is not None:  # Counter might have just one input
            # analog triggers are not supported for DigitalGlitchFilter
            if input.isAnalogTrigger():
                raise ValueError(
                    "Analog Triggers are not supported for DigitalGlitchFilters"
                )
            hal.setFilterSelect(input.getPortHandleForRouting(), channelIndex)

            selected = hal.getFilterSelect(input.getPortHandleForRouting())
            if selected != channelIndex:
                raise ValueError(
                    "setFilterSelect(%s) failed -> %s" % (channelIndex, selected)
                )

    def add(self, input: Union[DigitalSource, Encoder, Counter]) -> None:
        """
        Assigns the :class:`.DigitalSource`, :class:`.Encoder`, or
        :class:`counter.Counter` to this glitch filter.

        :param input: The object to add
        """
        if isinstance(input, DigitalSource):
            DigitalGlitchFilter._setFilter(input, self.channelIndex + 1)
        elif isinstance(input, Encoder):
            DigitalGlitchFilter._setFilter(input.aSource, self.channelIndex + 1)
            DigitalGlitchFilter._setFilter(input.bSource, self.channelIndex + 1)
        elif isinstance(input, Counter):
            DigitalGlitchFilter._setFilter(input.upSource, self.channelIndex + 1)
            DigitalGlitchFilter._setFilter(input.downSource, self.channelIndex + 1)
        else:
            raise ValueError("Cannot add %s to glitch filter" % input)

    def remove(self, input: Union[DigitalSource, Encoder, Counter]) -> None:
        """
        Removes this filter from the given input object
        """

        if isinstance(input, DigitalSource):
            DigitalGlitchFilter._setFilter(input, 0)
        elif isinstance(input, Encoder):
            DigitalGlitchFilter._setFilter(input.aSource, 0)
            DigitalGlitchFilter._setFilter(input.bSource, 0)
        elif isinstance(input, Counter):
            DigitalGlitchFilter._setFilter(input.upSource, 0)
            DigitalGlitchFilter._setFilter(input.downSource, 0)
        else:
            raise ValueError("Cannot remove %s from glitch filter" % input)

    def setPeriodCycles(self, fpga_cycles: int) -> None:
        """
        Sets the number of FPGA cycles that the input must hold steady to pass
        through this glitch filter.

        :param fpga_cycles: The number of FPGA cycles.
        """
        hal.setFilterPeriod(self.channelIndex, fpga_cycles)

    def setPeriodNanoSeconds(self, nanoseconds: float) -> None:
        """
        Sets the number of nanoseconds that the input must hold steady to pass
        through this glitch filter.

        :param nanoseconds: The number of nanoseconds.
        """
        fpga_cycles = int(
            nanoseconds * SensorUtil.kSystemClockTicksPerMicrosecond / 4 / 1000
        )
        self.setPeriodCycles(fpga_cycles)

    def getPeriodCycles(self) -> int:
        """
        Gets the number of FPGA cycles that the input must hold steady to pass
        through this glitch filter.

        :returns: The number of cycles.
        """
        return hal.getFilterPeriod(self.channelIndex)

    def getPeriodNanoSeconds(self) -> float:
        """
        Gets the number of nanoseconds that the input must hold steady to pass
        through this glitch filter.

        :returns: The number of nanoseconds.
        """
        fpga_cycles = self.getPeriodCycles()
        return fpga_cycles * 1000 / (SensorUtil.kSystemClockTicksPerMicrosecond / 4)

    def initSendable(self, builder: SendableBuilder) -> None:
        pass



# validated: 2018-11-18 EN 175c6c1f0130 edu/wpi/first/wpilibj/shuffleboard/ShuffleboardLayout.java
# ----------------------------------------------------------------------------
# Copyright (c) 2018 FIRST. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------
from networktables.networktable import NetworkTable
from .component import ShuffleboardComponent
from .container import ShuffleboardContainer


class ShuffleboardLayout(ShuffleboardComponent, ShuffleboardContainer):
    def __init__(self, parent: ShuffleboardContainer, name: str, type: str):
        ShuffleboardComponent.__init__(self, parent, name, type)
        ShuffleboardContainer.__init__(self)

    def buildInto(self, parentTable: NetworkTable, metaTable: NetworkTable) -> None:
        self.buildMetadata(metaTable)
        table = parentTable.getSubTable(self.getTitle())
        table.getEntry(".type").setString("ShuffleboardLayout")
        for component in self.getComponents():
            component.buildInto(table, metaTable.getSubTable(component.getTitle()))



# validated: 2018-11-17 EN 89d15f061b33 edu/wpi/first/wpilibj/smartdashboard/SmartDashboard.java

# validation note: 2017-10-22: Not using the getEntry() stuff that Java uses,
#                              as using the existing table stuff is more
#                              efficient

# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------
from typing import List, TypeVar, Union, Tuple, Iterable

T = TypeVar("T")

import hal
import threading

from networktables.entry import NetworkTableEntry
from networktables.networktable import NetworkTable

from .sendable import Sendable
from ._impl.utils import match_arglist, HasAttribute
from .sendablebuilder import SendableBuilder

__all__ = ["SmartDashboard"]


class Data:
    def __init__(self, sendable: Sendable) -> None:
        self.sendable = sendable
        self.builder = SendableBuilder()


class SmartDashboard:
    """The bridge between robot programs and the SmartDashboard on the laptop

    When a value is put into the SmartDashboard, it pops up on the
    SmartDashboard on the remote host. Users can put values into and get values
    from the SmartDashboard.
    
    These values can also be accessed by a NetworkTables client via the
    'SmartDashboard' table::
    
        from networktables import NetworkTables
        sd = NetworkTables.getTable('SmartDashboard')
        
        # sd.putXXX and sd.getXXX work as expected here
    
    """

    # The NetworkTable used by SmartDashboard
    table = None
    # A table linking tables in the SmartDashboard to the SmartDashboardData
    # objects they came from.
    tablesToData = {}
    mutex = threading.RLock()

    @classmethod
    def _reset(cls) -> None:
        cls.tablesToData = {}
        cls.table = None

    @classmethod
    def getTable(cls) -> NetworkTable:
        if cls.table is None:
            from networktables import NetworkTables

            cls.table = NetworkTables.getTable("SmartDashboard")
            hal.report(hal.UsageReporting.kResourceType_SmartDashboard, 0)
        return cls.table

    @classmethod
    def putData(cls, *args, **kwargs) -> None:
        """
        Maps the specified key (name of the :class:`.Sendable` if not provided) 
        to the specified value in this table.
        The value can be retrieved by calling the get method with a key that
        is equal to the original key.

        Two argument formats are supported: 
        
        - key, data
        - value
        
        :param key: the key (cannot be None)
        :type key: str
        :param data: the value
        :type data: :class:`.Sendable`
        :param value: the value
        :type value: :class:`.Sendable`
        """
        with cls.mutex:
            key_arg = ("key", [str])
            data_arg = ("data", [HasAttribute("initSendable")])
            value_arg = ("value", [HasAttribute("initSendable")])
            templates = [[key_arg, data_arg], [value_arg]]

            index, results = match_arglist(
                "SmartDashboard.putData", args, kwargs, templates
            )
            if index == 0:
                key = results["key"]
                data = results["data"]
            elif index == 1:
                data = results["value"]
                key = data.getName()
            else:
                raise ValueError("only (key, data) or (value) accepted")

            sddata = cls.tablesToData.get(key, None)
            if sddata is None or sddata.sendable != data:
                if sddata is not None:
                    sddata.builder.stopListeners()

                sddata = Data(data)
                cls.tablesToData[key] = sddata
                dataTable = cls.getTable().getSubTable(key)
                sddata.builder.setTable(dataTable)
                data.initSendable(sddata.builder)
                sddata.builder.updateTable()
                sddata.builder.startListeners()
                dataTable.getEntry(".name").setString(key)

    @classmethod
    def getData(cls, key: str) -> Sendable:
        """Returns the value at the specified key.
        
        :param key: the key

        :returns: the value
        
        :raises: :exc:`KeyError` if the key doesn't exist
        """
        data = cls.tablesToData.get(key, None)
        if data is None:
            raise KeyError("SmartDashboard data does not exist: '%s'" % key)
        return data.sendable

    @classmethod
    def getEntry(cls, key: str) -> NetworkTableEntry:
        """Gets the entry for the specified key.
        
        :param key: the key name
        """
        table = cls.getTable()
        return table.getEntry(key)

    @classmethod
    def containsKey(cls, key: str) -> bool:
        """Checks the table and tells if it contains the specified key.

        :param key: key the key to search for
        
        :returns: true if the table as a value assigned to the given key
        """
        table = cls.getTable()
        return table.containsKey(key)

    @classmethod
    def getKeys(cls, types: int = 0) -> List[str]:
        """Get the keys stored in the SmartDashboard table of NetworkTables.

        :param types: bitmask of types; 0 is treated as a "don't care".
        
        :returns: keys currently in the table
        """
        table = cls.getTable()
        return table.getKeys(types)

    @classmethod
    def setPersistent(cls, key: str) -> None:
        """Makes a key's value persistent through program restarts.
        The key cannot be null.

        :param key: the key name
        """
        table = cls.getTable()
        table.setPersistent(key)

    @classmethod
    def clearPersistent(cls, key: str) -> None:
        """Stop making a key's value persistent through program restarts.
        The key cannot be null.

        :param key: the key name
        """
        table = cls.getTable()
        table.clearPersistent(key)

    @classmethod
    def isPersistent(cls, key: str) -> bool:
        """Returns whether the value is persistent through program restarts.
        The key cannot be null.

        :param key: the key name
        
        :returns: True if the value is persistent.
        """
        table = cls.getTable()
        return table.isPersistent(key)

    @classmethod
    def setFlags(cls, key: str, flags: int) -> None:
        """Sets flags on the specified key in this table. The key can
        not be null.

        :param key: the key name
        :param flags: the flags to set (bitmask)
        """
        table = cls.getTable()
        table.setFlags(key, flags)

    @classmethod
    def clearFlags(cls, key: str, flags: int) -> None:
        """Clears flags on the specified key in this table. The key can
        not be null.

        :param key: the key name
        :param flags: the flags to clear (bitmask)
        """
        table = cls.getTable()
        table.clearFlags(key, flags)

    @classmethod
    def getFlags(cls, key: str) -> int:
        """ Returns the flags for the specified key.

        :param key: the key name
        
        :returns: the flags, or 0 if the key is not defined
        """
        table = cls.getTable()
        return table.getFlags(key)

    @classmethod
    def delete(cls, key: str) -> None:
        """Deletes the specified key in this table. The key can
        not be null.

        :param key: the key name
        """
        table = cls.getTable()
        table.delete(key)

    @classmethod
    def putBoolean(cls, key: str, value: bool) -> bool:
        """Put a boolean in the table.

        :param key: the key to be assigned to
        :param value: the value that will be assigned
        
        :return False if the table key already exists with a different type
        """
        table = cls.getTable()
        return table.putBoolean(key, value)

    @classmethod
    def setDefaultBoolean(cls, key: str, defaultValue: bool) -> bool:
        """Gets the current value in the table, setting it if it does not exist.
        
        :param key: the key
        :param defaultValue: the default value to set if key doens't exist.
        
        :returns: False if the table key exists with a different type
        """
        table = cls.getTable()
        return table.setDefaultBoolean(key, defaultValue)

    @classmethod
    def getBoolean(cls, key: str, defaultValue: T) -> Union[T, bool]:
        """Returns the boolean the key maps to. If the key does not exist or is of
        different type, it will return the default value.
        
        :param key: the key to look up
        :param defaultValue: returned if the key doesn't exist
        
        :returns: the value associated with the given key or the given default value
                  if there is no value associated with the key
        """
        table = cls.getTable()
        return table.getBoolean(key, defaultValue)

    @classmethod
    def putNumber(cls, key: str, value: float) -> bool:
        """Put a number in the table.
        
        :param key: the key to be assigned to
        :param value: the value that will be assigned
        
        :returns: False if the table key already exists with a different type
        """
        table = cls.getTable()
        return table.putNumber(key, value)

    @classmethod
    def setDefaultNumber(cls, key: str, defaultValue: float) -> bool:
        """Gets the current value in the table, setting it if it does not exist.
        
        :param key: the key
        :param defaultValue: the default value to set if key doens't exist.
        
        :returns: False if the table key exists with a different type
        """
        table = cls.getTable()
        return table.setDefaultNumber(key, defaultValue)

    @classmethod
    def getNumber(cls, key: str, defaultValue: T) -> Union[T, float]:
        """Returns the number the key maps to. If the key does not exist or is of
        different type, it will return the default value.
        
        :param key: the key to look up
        :param defaultValue: returned if the key doesn't exist
        
        :returns: the value associated with the given key or the given default value
                  if there is no value associated with the key
        """
        table = cls.getTable()
        return table.getNumber(key, defaultValue)

    @classmethod
    def putString(cls, key: str, value: str) -> bool:
        """Put a string in the table.
        
        :param key: the key to be assigned to
        :param value: the value that will be assigned
        
        :returns: False if the table key already exists with a different type
        """
        table = cls.getTable()
        return table.putString(key, value)

    @classmethod
    def setDefaultString(cls, key: str, defaultValue: str) -> bool:
        """Gets the current value in the table, setting it if it does not exist.
        
        :param key: the key
        :param defaultValue: the default value to set if key doens't exist.
        
        :returns: False if the table key exists with a different type
        """
        table = cls.getTable()
        return table.setDefaultString(key, defaultValue)

    @classmethod
    def getString(cls, key: str, defaultValue: T) -> Union[T, str]:
        """Returns the string the key maps to. If the key does not exist or is of
        different type, it will return the default value.
        
        :param key: the key to look up
        :param defaultValue: returned if the key doesn't exist
        
        :returns: the value associated with the given key or the given default value
                  if there is no value associated with the key
        """
        table = cls.getTable()
        return table.getString(key, defaultValue)

    @classmethod
    def putBooleanArray(cls, key: str, value: Iterable[bool]) -> bool:
        """Put a boolean array in the table.
        
        :param key: the key to be assigned to
        :param value: the value that will be assigned
        
        :returns: False if the table key already exists with a different type
        """
        table = cls.getTable()
        return table.putBooleanArray(key, value)

    @classmethod
    def setDefaultBooleanArray(cls, key: str, defaultValue: Iterable[bool]) -> bool:
        """Gets the current value in the table, setting it if it does not exist.
        
        :param key: the key
        :param defaultValue: the default value to set if key doens't exist.
        
        :returns: False if the table key exists with a different type
        """
        table = cls.getTable()
        return table.setDefaultBooleanArray(key, defaultValue)

    @classmethod
    def getBooleanArray(cls, key: str, defaultValue: T) -> Union[T, Tuple[bool, ...]]:
        """Returns the boolean array the key maps to. If the key does not exist or is of
        different type, it will return the default value.
        
        :param key: the key to look up
        :param defaultValue: returned if the key doesn't exist
        
        :returns: the value associated with the given key or the given default value
                  if there is no value associated with the key
        """
        table = cls.getTable()
        return table.getBooleanArray(key, defaultValue)

    @classmethod
    def putNumberArray(cls, key: str, value: Iterable[float]) -> bool:
        """Put a number array in the table.
        
        :param key: the key to be assigned to
        :param value: the value that will be assigned
        
        :returns: False if the table key already exists with a different type
        """
        table = cls.getTable()
        return table.putNumberArray(key, value)

    @classmethod
    def setDefaultNumberArray(cls, key: str, defaultValue: Iterable[float]) -> bool:
        """Gets the current value in the table, setting it if it does not exist.
        
        :param key: the key
        :param defaultValue: the default value to set if key doens't exist.
        
        :returns: False if the table key exists with a different type
        """
        table = cls.getTable()
        return table.setDefaultNumberArray(key, defaultValue)

    @classmethod
    def getNumberArray(cls, key: str, defaultValue: T) -> Union[T, Tuple[float, ...]]:
        """Returns the number array the key maps to. If the key does not exist or is of
        different type, it will return the default value.

        :param key: the key to look up
        :param defaultValue: returned if the key doesn't exist
        
        :returns: the value associated with the given key or the given default value
                  if there is no value associated with the key
        """
        table = cls.getTable()
        return table.getNumberArray(key, defaultValue)

    @classmethod
    def putStringArray(cls, key: str, value: Iterable[str]) -> bool:
        """Put a string array in the table
        
        :param key: the key to be assigned to
        :param value: the value that will be assigned

        :returns: False if the table key already exists with a different type
        """
        table = cls.getTable()
        return table.putStringArray(key, value)

    @classmethod
    def setDefaultStringArray(cls, key: str, defaultValue: Iterable[str]) -> bool:
        """If the key doesn't currently exist, then the specified value will
        be assigned to the key.
        
        :param key: the key to be assigned to
        :param defaultValue: the default value to set if key doesn't exist.

        :returns: False if the table key exists with a different type
        """
        table = cls.getTable()
        return table.setDefaultStringArray(key, defaultValue)

    @classmethod
    def getStringArray(cls, key: str, defaultValue: T) -> Union[T, Tuple[str, ...]]:
        """Returns the string array the key maps to. If the key does not exist or is
        of different type, it will return the default value.
        
        :param key: the key to look up
        :param defaultValue: the value to be returned if no value is found

        :returns: the value associated with the given key or the given default value
                  if there is no value associated with the key
        """
        table = cls.getTable()
        return table.getStringArray(key, defaultValue)

    @classmethod
    def putRaw(cls, key: str, value: bytes) -> bool:
        """Put a raw value (byte array) in the table.
        
        :param key: the key to be assigned to
        :param value: the value that will be assigned
        
        :returns: False if the table key already exists with a different type
        """
        table = cls.getTable()
        return table.putRaw(key, value)

    @classmethod
    def setDefaultRaw(cls, key: str, defaultValue: bytes) -> bool:
        """Gets the current value in the table, setting it if it does not exist.
        
        :param key: the key
        :param defaultValue: the default value to set if key doens't exist.
        
        :returns: False if the table key exists with a different type
        """
        table = cls.getTable()
        return table.setDefaultRaw(key, defaultValue)

    @classmethod
    def getRaw(cls, key: str, defaultValue: T) -> Union[T, bytes]:
        """Returns the raw value (byte array) the key maps to. If the key does not exist or is of
        different type, it will return the default value.

        :param key: the key to look up
        :param defaultValue: returned if the key doesn't exist
        
        :returns: the value associated with the given key or the given default value
                  if there is no value associated with the key
        """
        table = cls.getTable()
        return table.getRaw(key, defaultValue)

    @classmethod
    def updateValues(cls) -> None:
        with cls.mutex:
            for data in cls.tablesToData.values():
                data.builder.updateTable()



# validated: 2018-11-18 EN 81498e6af915 edu/wpi/first/wpilibj/IterativeRobot.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

import warnings
import hal
import logging

from .iterativerobotbase import IterativeRobotBase

__all__ = ["IterativeRobot"]


class IterativeRobot(IterativeRobotBase):
    """IterativeRobot implements the IterativeRobotBase robot program framework.

    The IterativeRobot class is intended to be subclassed by a user creating a robot program.

    periodic() functions from the base class are called each time a new packet is received from
    the driver station.

    .. deprecated:: 2019.0.0
        Use TimedRobot instead. It's a drop-in replacement that provides more
        regular execution periods.
    """

    kPacketPeriod = 0.02

    #: A python logging object that you can use to send messages to the log. It
    #: is recommended to use this instead of print statements.
    logger = logging.getLogger("robot")

    def __init__(self) -> None:
        """Constructor for IterativeRobot.

        The constructor initializes the instance variables for the robot to
        indicate the status of initialization for disabled, autonomous, and
        teleop code.

        .. warning:: If you override ``__init__`` in your robot class, you must call
                     the base class constructor. This must be used to ensure that
                     the communications code starts.
        """
        super().__init__(IterativeRobot.kPacketPeriod)
        hal.report(
            hal.UsageReporting.kResourceType_Framework,
            hal.UsageReporting.kFramework_Iterative,
        )

    def __init_subclass__(cls, **kwargs) -> None:
        warnings.warn(
            "IterativeRobot is deprecated, use TimedRobot instead.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        super().__init_subclass__(**kwargs)

    def startCompetition(self) -> None:
        """Provide an alternate "main loop" via startCompetition()."""

        self.robotInit()

        # Tell the DS that the robot is ready to be enabled
        hal.observeUserProgramStarting()

        # loop forever, calling the appropriate mode-dependent function
        while True:
            # Wait for new data to arrive
            self.ds.waitForData()
            # Call the appropriate function depending upon the current robot mode
            self.loopFunc()



# validated: 2018-01-04 TW 8b7aa61091df edu/wpi/first/wpilibj/ControllerPower.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

import hal

__all__ = ["ControllerPower"]


class ControllerPower:
    """Old Controller PR class.

    .. deprecated:: 2018.0.0
        Use RobotController class instead
    """

    @staticmethod
    def getInputVoltage() -> float:
        """
            Get the input voltage to the robot controller

            :returns: The controller input voltage value in Volts
        """
        return hal.getVinVoltage()

    @staticmethod
    def getInputCurrent() -> float:
        """
            Get the input current to the robot controller

            :returns: The controller input current value in Amps
        """
        return hal.getVinCurrent()

    @staticmethod
    def getVoltage3V3() -> float:
        """
            Get the voltage of the 3.3V rail

            :returns: The controller 3.3V rail voltage value in Volts
        """
        return hal.getUserVoltage3V3()

    @staticmethod
    def getCurrent3V3() -> float:
        """
            Get the current output of the 3.3V rail

            :returns: The controller 3.3V rail output current value in Amps
        """
        return hal.getUserCurrent3V3()

    @staticmethod
    def getEnabled3V3() -> bool:
        """
            Get the enabled state of the 3.3V rail. The rail may be
            disabled due to a controller brownout, a short circuit on the
            rail, or controller over-voltage

            :returns: True if enabled, False otherwise
        """
        return hal.getUserActive3V3()

    @staticmethod
    def getFaultCount3V3() -> int:
        """
            Get the count of the total current faults on the 3.3V rail since
            the controller has booted

            :returns: The number of faults
        """
        return hal.getUserCurrentFaults3V3()

    @staticmethod
    def getVoltage5V() -> float:
        """
            Get the voltage of the 5V rail


            :returns: The controller 5V rail voltage value in Volts
        """
        return hal.getUserVoltage5V()

    @staticmethod
    def getCurrent5V() -> float:
        """
            Get the current output of the 5V rail

            :returns: The controller 5V rail output current value in Amps
        """
        return hal.getUserCurrent5V()

    @staticmethod
    def getEnabled5V() -> bool:
        """
            Get the enabled state of the 5V rail. The rail may be disabled
            due to a controller brownout, a short circuit on the rail, or
            controller over-voltage

            :returns: True if enabled, False otherwise
        """
        return hal.getUserActive5V()

    @staticmethod
    def getFaultCount5V() -> int:
        """
            Get the count of the total current faults on the 5V rail since
            the controller has booted

            :returns: The number of faults
        """
        return hal.getUserCurrentFaults5V()

    @staticmethod
    def getVoltage6V() -> float:
        """
            Get the voltage of the 6V rail

            :returns: The controller 6V rail voltage value in Volts
        """
        return hal.getUserVoltage6V()

    @staticmethod
    def getCurrent6V() -> float:
        """
            Get the current output of the 6V rail

            :returns: The controller 6V rail output current value in Amps
        """
        return hal.getUserCurrent6V()

    @staticmethod
    def getEnabled6V() -> bool:
        """
            Get the enabled state of the 6V rail. The rail may be disabled
            due to a controller brownout, a short circuit on the rail, or
            controller over-voltage

            :returns: True if enabled, False otherwise
        """
        return hal.getUserActive6V()

    @staticmethod
    def getFaultCount6V() -> int:
        """
            Get the count of the total current faults on the 6V rail since
            the controller has booted

            :returns: The number of faults
        """
        return hal.getUserCurrentFaults6V()



# validated: 2018-12-15 EN 6f0c185a05c9 edu/wpi/first/wpilibj/shuffleboard/ShuffleboardRoot.java
# ----------------------------------------------------------------------------
# Copyright (c) 2018 FIRST. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .tab import ShuffleboardTab


class ShuffleboardRoot:
    """
    The root of the data placed in Shuffleboard. It contains the tabs, but no 
    data is placed directly in the root.
    """

    def getTab(self, title: str) -> "ShuffleboardTab":
        """
        Gets the tab with the given title, creating it if it does not 
        already exist.

        :param title: the title of the tab
        :returns: the tab with the given title
        """
        raise NotImplementedError

    def update(self) -> None:
        """Updates all tabs."""
        raise NotImplementedError

    def enableActuatorWidgets(self) -> None:
        """Enables all widgets in Shuffleboard that offer user control over actuators."""
        raise NotImplementedError

    def disableActuatorWidgets(self) -> None:
        """Disables all widgets in Shuffleboard that offer user control over actuators."""
        raise NotImplementedError

    def selectTab(self, index_or_title) -> None:
        """
        Selects the tab in the dashboard with the given index in the 
        range [0..n-1], where `n` is the number of tabs in the dashboard at 
        the time this method is called.

        Or

        Selects the tab in the dashboard with the given title.

        :param index_or_title: when integer, the index of the tab to select.
                               when string, the title of the tab to select.
        """
        raise NotImplementedError



"""
    Classes in this package are used to interface various types of buttons
    to a command-based robot.
    
    If you are not using the Command framework, you can ignore these classes.
"""

from .button import Button
from .internalbutton import InternalButton
from .joystickbutton import JoystickButton
from .networkbutton import NetworkButton
from .trigger import Trigger



# validated: 2018-11-18 EN 175c6c1f0130 edu/wpi/first/wpilibj/shuffleboard/SimpleWidget.java
# ----------------------------------------------------------------------------
# Copyright (c) 2018 FIRST. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------
from networktables.entry import NetworkTableEntry
from .widget import ShuffleboardWidget
from .container import ShuffleboardContainer
from .layout import ShuffleboardLayout


class SimpleWidget(ShuffleboardWidget):
    """A Shuffleboard widget that handles a single data point such as a number or string."""

    def __init__(self, parent: ShuffleboardContainer, title: str):
        super().__init__(parent, title)
        self.entry = None

    def getEntry(self) -> NetworkTableEntry:
        """Gets the NetworkTable entry that contains the data for this widget."""
        if self.entry is None:
            self._forceGenerate()

        return self.entry

    def buildInto(self, parentTable, metaTable) -> None:
        self.buildMetadata(metaTable)
        if self.entry is None:
            self.entry = parentTable.getEntry(self.getTitle())

    def _forceGenerate(self):
        parent = self.getParent()  # type: ShuffleboardContainer
        while isinstance(parent, ShuffleboardLayout):
            parent = parent.getParent()

        tab = parent  # type: ShuffleboardTab
        tab.getRoot().update()



# validated: 2018-11-08 EN e2100730447d edu/wpi/first/wpilibj/Notifier.java
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2016-2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

import threading

import hal

from .resource import Resource
from .robotcontroller import RobotController


class Notifier:
    def __init__(self, run: callable) -> None:
        """
        Create a Notifier for timer event notification.

        :param run: The handler that is called at the notification time which is
                    set using :meth:`.startSingle` or :meth:`.startPeriodic`.
        """
        #: The lock for the process information.
        self._processLock = threading.RLock()

        # Notifier handle
        self._notifier = hal.initializeNotifier()

        #: The time, in microseconds, at which the corresponding handler should be
        #: called. Has the same zero as Timer.getFPGATime().
        self._expirationTime = 0

        #: The handler passed in by the user which should be called at the
        #: appropriate interval.
        self._handler = run

        # Whether we are calling the handler just once or periodically.
        self._periodic = False

        #: If periodic, the period of the calling; if just once, stores how long it
        #: is until we call the handler.
        self._period = 0

        #: The thread waiting on the HAL alarm
        self._thread = threading.Thread(target=self._run, name="Notifier", daemon=True)
        self._thread.start()

        # python-specific
        Resource._add_global_resource(self)

    def close(self) -> None:
        handle, self._notifier = self._notifier, None
        if not handle:
            return

        hal.stopNotifier(handle)

        # Join the thread to ensure the handler has exited.
        if self._thread.is_alive():
            # python-specific: interrupt not supported
            self._thread.join()

        hal.cleanNotifier(handle)
        self._thread = None

    def _updateAlarm(self, triggerTime=None) -> None:
        """
        Update the alarm hardware to reflect the next alarm.

        :param triggerTime: the time at which the next alarm will be triggered
        """
        if triggerTime is None:
            triggerTime = int(self._expirationTime * 1e6)
        handle = self._notifier
        if handle:
            hal.updateNotifierAlarm(handle, triggerTime)

    def _run(self) -> None:
        while True:
            notifier = self._notifier
            if not notifier:
                break

            curTime = hal.waitForNotifierAlarm(notifier)
            if curTime == 0:
                break

            with self._processLock:
                handler = self._handler
                if self._periodic:
                    self._expirationTime += self._period
                    self._updateAlarm()
                else:
                    # need to update the alarm to cause it to wait again
                    self._updateAlarm(-1)

            if handler:
                handler()

    def setHandler(self, handler: callable) -> None:
        """Change the handler function.
        
        :param handler: Handler
        """
        with self._processLock:
            self._handler = handler

    def startSingle(self, delay: float) -> None:
        """Register for single event notification.

        A timer event is queued for a single event after the specified delay.

        :param delay: Seconds to wait before the handler is called.
        """
        with self._processLock:
            self._periodic = False
            self._period = delay
            self._expirationTime = RobotController.getFPGATime() * 1e-6 + delay
            self._updateAlarm()

    def startPeriodic(self, period: float) -> None:
        """Register for periodic event notification.

        A timer event is queued for periodic event notification.
        Each time the interrupt occurs, the event will be immediately
        requeued for the same time interval.

        :param period: Period in seconds to call the handler starting
                       one period after the call to this method.
        """
        with self._processLock:
            self._periodic = True
            self._period = period
            self._expirationTime = RobotController.getFPGATime() * 1e-6 + period
            self._updateAlarm()

    def stop(self) -> None:
        """Stop timer events from occurring.

        Stop any repeating timer events from occurring. This will also
        remove any single notification events from the queue.
        If a timer-based call to the registered handler is in progress,
        this function will block until the handler call is complete.
        """
        hal.cancelNotifierAlarm(self._notifier)
